
#if !defined(NDEBUG)
#define _CRTDBG_MAP_ALLOC
#include <crtdbg.h>
#endif

#include <SAPS/SupportLib.h>
#include <assert.h>
#include <math.h>
#include <windows.h>

#pragma warning(disable: 4996)


static bool Linear_TextToValue(FloatResponse const *resp, wchar_t const *text, float *oValue)
{
  wchar_t *oot = 0;
  *oValue = (float)wcstod(text, &oot);
  return oot != text && oot != 0;
}

static bool Linear_ValueToText(FloatResponse const *resp, float iValue, wchar_t *oText, size32 count)
{
  swprintf(oText, count, L"%f", iValue);
  oText[count-1] = 0;
  return true;
}

static float Linear_WarpFromGuiToValue(FloatResponse const *resp, float guiValue, float minValue, float maxValue)
{
  return guiValue * (maxValue - minValue) + minValue;
}

static float Linear_WarpFromValueToGui(FloatResponse const *resp, float value, float minValue, float maxValue)
{
  if (maxValue == minValue) return 1;
  return (value - minValue) / (maxValue - minValue);
}

static FloatResponse const rLinear = {
  sizeof(FloatResponse),
  &Linear_TextToValue,
  &Linear_ValueToText,
  &Linear_WarpFromGuiToValue,
  &Linear_WarpFromValueToGui,
};



static const float linToDbConversion = 20.0f / logf(10.0f);
static const float dBToLinConversion = logf(10.0f) / 20.0f;

static bool dB_TextToValue(FloatResponse const *resp, wchar_t const *text, float *oValue)
{
  while (*text && iswspace(*text))
    ++text;
  if (!wcsicmp(text, L"-oo") || !wcsicmp(text, L"-Inf") || !wcsicmp(text, L"-00"))
  {
    *oValue = 0;
    return true;
  }
  wchar_t *oot = 0;
  *oValue = (float)wcstod(text, &oot);
  if (oot == text || oot == 0)
    return false;
  *oValue = expf(*oValue * dBToLinConversion);
  return true;
}

static bool dB_ValueToText(FloatResponse const *resp, float iValue, wchar_t *oText, size32 count)
{
  if (iValue < 0) return false;
  if (iValue == 0)
  {
    swprintf(oText, count, L"-oo");
    oText[count-1] = 0;
    return true;
  }
  float pv = logf(iValue) * linToDbConversion;
  swprintf(oText, count, L"%.1f", pv);
  oText[count-1] = 0;
  return true;
}

static float dB_WarpFromGuiToValue(FloatResponse const *resp, float guiValue, float minValue, float maxValue)
{
  return guiValue * guiValue * (maxValue - minValue) + minValue;
}

static float dB_WarpFromValueToGui(FloatResponse const *resp, float value, float minValue, float maxValue)
{
  if (maxValue == minValue) return 1;
  return sqrtf((value - minValue) / (maxValue - minValue));
}

static FloatResponse const rDb = {
  sizeof(FloatResponse),
  &dB_TextToValue,
  &dB_ValueToText,
  &dB_WarpFromGuiToValue,
  &dB_WarpFromValueToGui,
};

FloatResponse const &LinearResponse()
{
  return rLinear;
}

FloatResponse const &DbResponse()
{
  return rDb;
}


static bool Selector_TextToValue(IntResponse const *resp, wchar_t const *text, size64 *oValue)
{
  bool ret = false;
  wchar_t *oot;
  *oValue = wcstol(text, &oot, 10);
  for (size_t i = 0; i < resp->ResponseCount; ++i)
  {
    if (!wcscmp(text, ((wchar_t const *const *)resp->ResponseData)[i]))
    {
      *oValue = i;
      return true;
    }
    if (!wcsicmp(text, ((wchar_t const *const *)resp->ResponseData)[i]))
    {
      *oValue = i;
      ret = true;
    }
  }
  return ret;
}

static bool Selector_ValueToText(IntResponse const *resp, size64 iValue, wchar_t *oText, size32 count)
{
  if (iValue < 0 || iValue >= resp->ResponseCount)
  {
    swprintf(oText, count, L"%d", iValue);
    oText[count-1] = 0;
    return false;
  }
  swprintf(oText, count, L"%s", ((wchar_t const *const *)resp->ResponseData)[(size_t)iValue]);
  oText[count-1] = 0;
  return true;
}

IntResponse &SelectorIntResponse(IntResponse *iResponse, wchar_t const *const *Options, size32 Count)
{
  iResponse->StructSize = sizeof(IntResponse);
  iResponse->ValueToText = &Selector_ValueToText;
  iResponse->TextToValue = &Selector_TextToValue;
  iResponse->ResponseData = (void *)Options;
  iResponse->ResponseCount = Count;
  return *iResponse;
}

static IntResponse rOnOff;
static wchar_t const *aOnOff[] = {
  L"Off", L"On",
};

IntResponse &OnOffIntResponse()
{
  return SelectorIntResponse(&rOnOff, aOnOff, 2);
}

static bool LinearInt_TextToValue(IntResponse const *resp, wchar_t const *text, size64 *oValue)
{
  wchar_t *oot = 0;
  *oValue = wcstol(text, &oot, 10);
  return (oot != 0 && oot != text);
}

static bool LinearInt_ValueToText(IntResponse const *resp, size64 iValue, wchar_t *oText, size32 count)
{
  swprintf(oText, count, L"%Ld", iValue);
  oText[count-1] = 0;
  return true;
}

static struct IntResponse rLinearInt =
{
  sizeof(IntResponse),
  &LinearInt_TextToValue,
  &LinearInt_ValueToText,
};

IntResponse &LinearIntResponse()
{
  return rLinearInt;
}





ParameterImpl::ParameterImpl(EffectBase *effect, void *userRef)
  : effect_(effect)
  , userRef_(userRef)
{
  effect_->host_->AllocateAddress(&address_, this, effect);
}

ParameterImpl::~ParameterImpl()
{
  effect_->host_->DeallocateAddress(address_);
}




FloatParameter::FloatParameter(FloatRange const &range, wchar_t const *name, wchar_t const *unit, EffectBase *effect, void *userRef)
  : ParameterImpl(effect, userRef)
  , range_(range)
  , name_(name)
  , unit_(unit)
  , value_(range.From)
{
}

Address FloatParameter::TargetAddress()
{
  return address_;
}

ParameterType FloatParameter::Type()
{
  return ParameterTypeFloat;
}

void const *FloatParameter::Value()
{
  return &value_;
}

size64 FloatParameter::Size()
{
  return sizeof(value_);
}

void FloatParameter::SetValue(void const *value, size64 size)
{
  if (size != sizeof(value_))
  {
    ::OutputDebugString(L"Ill behaved host: FloatParameter::SetValue() with bad size.\r\n");
    return;
  }
  value_ = *(float *)value;
}

float FloatParameter::Read(size32 atOffset)
{
  return value_;
}

